from pprint import pformat
import os
import shutil
from StringIO import StringIO
import tempfile
import yaml

from testtools import TestCase
from mock import patch, call, MagicMock

import hooks


FIXTURES = os.path.join(os.path.dirname(__file__), 'fixtures')


original_open = open


class BalancerRelationTest(TestCase):

    def setUp(self):
        super(BalancerRelationTest, self).setUp()

        self.get_relation_data = self.patch_hook("get_relation_data")
        self.config_get = self.patch_hook("config_get")
        self.log = self.patch_hook("log")
        self.write_balancer_config = self.patch_hook("write_balancer_config")

    def patch_hook(self, hook_name):
        mock_controller = patch.object(hooks, hook_name)
        mock = mock_controller.start()
        self.addCleanup(mock_controller.stop)
        return mock

    def test_relation_data_returns_none(self):
        self.get_relation_data.return_value = None
        self.assertIs(None, hooks.update_balancers())
        self.log.assert_called_once_with("No relation data, exiting.")
        self.write_balancer_config.assert_not_called()

    def test_relation_data_returns_no_relations(self):
        self.get_relation_data.return_value = ()
        self.assertIs(None, hooks.update_balancers())
        self.log.assert_called_once_with("No relation data, exiting.")
        self.write_balancer_config.assert_not_called()

    def test_no_port_in_relation_data(self):
        self.get_relation_data.return_value = {
            "foo": {"private-address": "1.2.3.4"},
        }
        self.assertIs(None, hooks.update_balancers())
        self.log.assert_called_once_with(
            "No port in relation data for 'foo', skipping.")
        self.write_balancer_config.assert_not_called()

    def test_no_private_address_in_relation_data(self):
        self.get_relation_data.return_value = {
            "foo": {"port": 4242},
        }
        self.assertIs(None, hooks.update_balancers())
        self.log.assert_called_once_with(
            "No private-address in relation data for 'foo', skipping.")
        self.write_balancer_config.assert_not_called()

    def test_sitenames_in_relation_data(self):
        self.get_relation_data.return_value = {
            "foo": {"private-address": "1.2.3.4",
                    "port": 4242,
                    "sitenames": "foo.internal bar.internal"},
        }
        expected = {
            "foo.internal": ["1.2.3.4:4242"],
            "bar.internal": ["1.2.3.4:4242"],
        }
        self.assertEqual(hooks.update_balancers(), expected)
        self.write_balancer_config.assert_called_once_with(expected)

    def test_all_services_in_relation_data(self):
        self.get_relation_data.return_value = {
            "foo": {
                "private-address": "1.2.3.4",
                "port": 80,
                "all_services": yaml.dump(
                    [
                        {"service_name": "foo.internal",
                         "service_port": 4242},
                        {"service_name": "bar.internal",
                         "service_port": 4243},
                    ]
                ),
            },
        }
        expected = {
            "foo.internal": ["1.2.3.4:4242"],
            "bar.internal": ["1.2.3.4:4243"],
        }
        self.assertEqual(hooks.update_balancers(), expected)
        self.write_balancer_config.assert_called_once_with(expected)

    def test_unit_names_in_relation_data(self):
        self.get_relation_data.return_value = {
            "foo-1": {"private-address": "1.2.3.4",
                      "port": 4242},
            "foo-2": {"private-address": "1.2.3.5",
                      "port": 4242},
        }
        expected = {
            "foo": ["1.2.3.4:4242", "1.2.3.5:4242"],
        }
        self.assertEqual(hooks.update_balancers(), expected)
        self.write_balancer_config.assert_called_once_with(expected)


class HelpersTest(TestCase):
    def setUp(self):
        super(HelpersTest, self).setUp()

        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        super(HelpersTest, self).tearDown()

        shutil.rmtree(self.tempdir, ignore_errors=True)

    def read_fixture(self, basename):
        fixture = os.path.join(FIXTURES, basename)
        with original_open(fixture) as f:
            content = f.read()
        return content

    @patch('subprocess.call')
    def test_opens_port(self, mock_call):
        mock_call.return_value = 'some result'

        result = hooks.open_port(port=123, protocol='icmp')

        self.assertEqual(result, 'some result')
        mock_call.assert_called_with(['open-port', "123/icmp"])

    @patch('subprocess.call')
    def test_opens_tcp_port_by_default(self, mock_call):
        mock_call.return_value = 'some result'

        result = hooks.open_port(port=123)

        self.assertEqual(result, 'some result')
        mock_call.assert_called_with(['open-port', "123/TCP"])

    @patch('subprocess.call')
    def test_doesnt_open_if_port_not_provided(self, mock_call):
        result = hooks.open_port()

        self.assertIsNone(result)
        self.assertFalse(mock_call.called)

    @patch('subprocess.call')
    def test_closes_port(self, mock_call):
        mock_call.return_value = 'some result'

        result = hooks.close_port(port=123, protocol='icmp')

        self.assertEqual(result, 'some result')
        mock_call.assert_called_with(['close-port', "123/icmp"])

    @patch('subprocess.call')
    def test_closes_tcp_port_by_default(self, mock_call):
        mock_call.return_value = 'some result'

        result = hooks.close_port(port=123)

        self.assertEqual(result, 'some result')
        mock_call.assert_called_with(['close-port', "123/TCP"])

    @patch('subprocess.call')
    def test_doesnt_close_if_port_not_provided(self, mock_call):
        result = hooks.close_port()

        self.assertIsNone(result)
        self.assertFalse(mock_call.called)

    @patch('subprocess.check_output')
    @patch('hooks.run')
    def test_gets_config(self, run, check_output):
        json_string = '{"foo": "BAR"}'
        check_output.return_value = json_string
        run.return_value = 'some-server-name'

        result = hooks.config_get()

        self.assertEqual(result, {
            'servername': 'some-server-name',
            u'foo': u'BAR',
        })
        check_output.assert_called_with(['config-get', '--format=json'])

    @patch('subprocess.check_output')
    @patch('hooks.run')
    def test_gets_config_with_scope(self, run, check_output):
        json_string = '{"foo": "BAR"}'
        check_output.return_value = json_string
        run.return_value = 'some-server-name'

        result = hooks.config_get(scope='baz')

        self.assertEqual(result, {
            'servername': 'some-server-name',
            u'foo': u'BAR',
        })
        check_output.assert_called_with(['config-get', 'baz', '--format=json'])

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_logs_and_returns_none_if_config_get_fails(self, log,
                                                       check_output):
        check_output.side_effect = RuntimeError('some error')

        result = hooks.config_get()

        log.assert_called_with('some error')
        self.assertIsNone(result)

    @patch('subprocess.call')
    def test_installs_packages(self, mock_call):
        mock_call.return_value = 'some result'

        result = hooks.apt_get_install('foo bar')

        self.assertEqual(result, 'some result')
        mock_call.assert_called_with(['apt-get', '-y', 'install', '-qq',
                                      'foo bar'])

    @patch('subprocess.call')
    def test_installs_nothing_if_package_not_provided(self, mock_call):
        self.assertFalse(hooks.apt_get_install())
        self.assertFalse(mock_call.called)

    @patch('subprocess.call')
    def test_purges_packages(self, mock_call):
        mock_call.return_value = 'some result'

        result = hooks.apt_get_purge('foo bar')

        self.assertEqual(result, 'some result')
        mock_call.assert_called_with(['apt-get', '-y', 'purge', '-qq',
                                      'foo bar'])

    @patch('subprocess.call')
    def test_purges_nothing_if_package_not_provided(self, mock_call):
        self.assertFalse(hooks.apt_get_purge())
        self.assertFalse(mock_call.called)

    @patch('subprocess.call')
    def test_starts_apache_successfully(self, mock_call):
        mock_call.return_value = 0
        action = 'start'

        self.assertTrue(hooks.service_apache2(action))
        mock_call.assert_called_with(['service', 'apache2', action])

    @patch('subprocess.call')
    def test_fails_to_start_apache(self, mock_call):
        mock_call.return_value = 1
        action = 'start'

        self.assertFalse(hooks.service_apache2(action))
        mock_call.assert_called_with(['service', 'apache2', action])

    @patch('subprocess.call')
    def test_checks_apache_successfully(self, mock_call):
        mock_call.return_value = 0

        self.assertTrue(hooks.service_apache2('check'))
        mock_call.assert_called_with(['/usr/sbin/apache2ctl', 'configtest'])

    @patch('subprocess.call')
    def test_fails_to_check_apache(self, mock_call):
        mock_call.return_value = 1

        self.assertFalse(hooks.service_apache2('check'))
        mock_call.assert_called_with(['/usr/sbin/apache2ctl', 'configtest'])

    @patch('subprocess.call')
    def test_fails_to_check_apache_with_another_return_value(self, mock_call):
        mock_call.return_value = 2

        self.assertFalse(hooks.service_apache2('check'))
        mock_call.assert_called_with(['/usr/sbin/apache2ctl', 'configtest'])

    @patch('subprocess.call')
    def test_does_nothing_if_action_not_provided(self, mock_call):
        self.assertIsNone(hooks.service_apache2())
        self.assertFalse(mock_call.called)

    @patch('subprocess.check_output')
    def test_runs_an_arbitrary_command(self, check_output):
        check_output.return_value = 'some result'

        self.assertEqual(hooks.run('foo', 1, 2, bar='baz'), 'some result')
        check_output.assert_called_with('foo', 1, 2, bar='baz')

    @patch('subprocess.check_output')
    @patch('sys.stdout.write')
    def test_prints_and_reraises_run_error(self, write, check_output):
        check_output.side_effect = RuntimeError('some error')

        self.assertRaises(RuntimeError, hooks.run, 'some command')
        self.assertEqual(write.mock_calls, [
            call('some error'),
            call('\n'),
        ])

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    @patch('hooks.apt_get_install')
    def test_enables_a_module(self, apt_get_install, path_exists,
                              service_apache2, log, mock_call):
        module = 'foo'
        module_already_enabled = False
        module_available = False
        apache_check = True
        apache_reload = None
        module_installed = 0
        module_finally_enabled = 0

        path_exists.side_effect = [module_already_enabled, module_available]
        service_apache2.side_effect = [apache_check, apache_reload]
        apt_get_install.return_value = module_installed
        mock_call.return_value = module_finally_enabled

        result = hooks.enable_module(module)

        self.assertTrue(result)
        path_exists.assert_has_calls([
            call("/etc/apache2/mods-enabled/%s.load" % (module)),
            call("/etc/apache2/mods-available/%s.load" % (module))
        ])
        apt_get_install.assert_called_with("libapache2-mod-%s" % (module))
        mock_call.assert_called_with(['/usr/sbin/a2enmod', module])
        service_apache2.assert_has_calls([call('check'), call('reload')])
        self.assertFalse(log.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    @patch('hooks.apt_get_install')
    def test_doesnt_enable_if_module_not_provided(self, apt_get_install,
                                                  path_exists, service_apache2,
                                                  log, mock_call):
        module = None

        result = hooks.enable_module(module)

        self.assertTrue(result)
        self.assertFalse(apt_get_install.called)
        self.assertFalse(path_exists.called)
        self.assertFalse(service_apache2.called)
        self.assertFalse(log.called)
        self.assertFalse(mock_call.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    @patch('hooks.apt_get_install')
    def test_doesnt_enable_if_module_already_enabled(self, apt_get_install,
                                                     path_exists,
                                                     service_apache2, log,
                                                     mock_call):
        module = 'foo'
        module_already_enabled = True

        path_exists.return_value = module_already_enabled

        result = hooks.enable_module(module)

        self.assertTrue(result)
        path_exists.assert_called_with(
            "/etc/apache2/mods-enabled/%s.load" % (module))
        log.assert_called_with("Module already loaded")
        self.assertFalse(apt_get_install.called)
        self.assertFalse(service_apache2.called)
        self.assertFalse(mock_call.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    @patch('hooks.apt_get_install')
    def test_fails_to_enable_if_module_not_installed(self, apt_get_install,
                                                     path_exists,
                                                     service_apache2, log,
                                                     mock_call):
        module = 'foo'
        module_already_enabled = False
        module_available = False
        module_installed = 1

        path_exists.side_effect = [module_already_enabled, module_available]
        apt_get_install.return_value = module_installed

        result = hooks.enable_module(module)

        self.assertFalse(result)
        path_exists.assert_has_calls([
            call("/etc/apache2/mods-enabled/%s.load" % (module)),
            call("/etc/apache2/mods-available/%s.load" % (module))
        ])
        apt_get_install.assert_called_with("libapache2-mod-%s" % (module))
        log.assert_called_with("Installing module %s failed" % (module))
        self.assertFalse(mock_call.called)
        self.assertFalse(service_apache2.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    @patch('hooks.apt_get_install')
    def test_fails_to_enable_if_enmod_fails(self, apt_get_install,
                                            path_exists, service_apache2,
                                            log, mock_call):
        module = 'foo'
        module_already_enabled = False
        module_available = False
        module_installed = 0
        module_finally_enabled = 1

        path_exists.side_effect = [module_already_enabled, module_available]
        apt_get_install.return_value = module_installed
        mock_call.return_value = module_finally_enabled

        result = hooks.enable_module(module)

        self.assertFalse(result)
        path_exists.assert_has_calls([
            call("/etc/apache2/mods-enabled/%s.load" % (module)),
            call("/etc/apache2/mods-available/%s.load" % (module))
        ])
        apt_get_install.assert_called_with("libapache2-mod-%s" % (module))
        mock_call.assert_called_with(['/usr/sbin/a2enmod', module])
        self.assertFalse(log.called)
        self.assertFalse(service_apache2.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    def test_disables_a_module(self, path_exists, service_apache2, log,
                               mock_call):
        module = 'foo'
        module_still_enabled = True
        apache_check = True
        apache_reload = None
        module_finally_disabled = 0

        path_exists.return_value = module_still_enabled
        service_apache2.side_effect = [apache_check, apache_reload]
        mock_call.return_value = module_finally_disabled

        result = hooks.disable_module(module)

        self.assertTrue(result)
        path_exists.assert_called_with(
            "/etc/apache2/mods-enabled/%s.load" % (module))
        mock_call.assert_called_with(['/usr/sbin/a2dismod', module])
        service_apache2.assert_has_calls([call('check'), call('reload')])
        self.assertFalse(log.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    def test_doest_disable_if_module_not_provided(self, path_exists,
                                                  service_apache2, log,
                                                  mock_call):
        module = None

        result = hooks.disable_module(module)

        self.assertTrue(result)
        self.assertFalse(path_exists.called)
        self.assertFalse(service_apache2.called)
        self.assertFalse(log.called)
        self.assertFalse(mock_call.called)

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    def test_does_nothing_if_module_already_disabled(self, path_exists,
                                                     service_apache2, log,
                                                     mock_call):
        module = 'foo'
        module_still_enabled = False

        path_exists.return_value = module_still_enabled

        result = hooks.disable_module(module)

        self.assertTrue(result)
        path_exists.assert_called_with(
            "/etc/apache2/mods-enabled/%s.load" % (module))
        log.assert_called_with("Module already disabled")

    @patch('subprocess.call')
    @patch('hooks.log')
    @patch('hooks.service_apache2')
    @patch('os.path.exists')
    def test_fails_to_disable_if_dismod_fails(self, path_exists,
                                              service_apache2, log, mock_call):
        module = 'foo'
        module_still_enabled = True
        apache_check = True
        apache_reload = None
        module_finally_disabled = 1

        path_exists.return_value = module_still_enabled
        service_apache2.side_effect = [apache_check, apache_reload]
        mock_call.return_value = module_finally_disabled

        result = hooks.disable_module(module)

        self.assertFalse(result)
        path_exists.assert_called_with(
            "/etc/apache2/mods-enabled/%s.load" % (module))
        mock_call.assert_called_with(['/usr/sbin/a2dismod', module])
        self.assertFalse(service_apache2.called)
        self.assertFalse(log.called)

    @patch('hooks.config_get')
    @patch('hooks.log')
    def test_writes_balancer_config(self, log, config_get):
        config_get.return_value = {
            'lb_balancer_timeout': 123,
        }
        balancer_config = {
            'foo': ['10.11.12.13'],
            'bar': ['10.11.12.14', '10.11.12.15'],
        }
        with patch('hooks.default_apache2_config_dir', self.tempdir):
            hooks.write_balancer_config(balancer_config)
        for balancer in balancer_config.keys():
            basename = '%s.balancer' % balancer
            exp_path = os.path.join(FIXTURES, basename)
            res_path = os.path.join(self.tempdir, basename)
            with open(exp_path) as exp, open(res_path) as res:
                self.assertEqual(exp.read(), res.read())

    @patch('hooks.config_get')
    @patch('pwd.getpwnam')
    @patch('grp.getgrnam')
    @patch('hooks.os')  # patch the reference that "hooks" has to "os"
    @patch('hooks.log')
    @patch('__builtin__.open')
    @patch('subprocess.call')
    def test_updates_nrpe_checks(self, call_, open_, log, os_, getgrnam,
                                 getpwnam, config_get):
        config_data = {
            'nagios_check_http_params': ['foo', 'bar'],
            'nagios_context': 'somecontext',
        }
        os_.environ = {
            'JUJU_UNIT_NAME': 'some/unit',
            'CHARM_DIR': os.environ['CHARM_DIR'],
        }
        nagios_logdir_exists = False
        nagios_exportdir_exists = True
        nrpe_service_config = MagicMock()
        nrpe_check_config = MagicMock()
        nrpe_service_template_path = os.path.join(
            os.environ['CHARM_DIR'], 'data', 'nrpe_service.template')
        with original_open(nrpe_service_template_path) as f:
            nrpe_service_template = StringIO(f.read())
        mock_file = MagicMock()
        nagios_server_exists = True
        nagios_uid = 1234
        nagios_gid = 2345
        nagios_logdir = '/var/log/nagios'
        nagios_exportdir = '/var/lib/nagios/export'

        config_get.return_value = config_data
        getpwnam('nagios').pw_uid = nagios_uid
        getgrnam('nagios').gr_gid = nagios_gid
        os_.path.exists.side_effect = [
            nagios_logdir_exists,
            nagios_exportdir_exists,
        ]
        os_.listdir.return_value = [
            'foo-check_vhost.cfg',
            'bar-check_vhost.cfg',
            'some-other-file.cfg'
        ]
        mock_file.__enter__.side_effect = [nrpe_service_config,
                                           nrpe_check_config]
        open_.side_effect = [
            nrpe_service_template,
            mock_file,
            mock_file,
        ]
        os_.path.isfile.return_value = nagios_server_exists
        os_.path.join = os.path.join

        hooks.update_nrpe_checks()

        getpwnam.assert_called_with('nagios')
        getgrnam.assert_called_with('nagios')
        self.assertEqual(os_.path.exists.mock_calls, [
            call(nagios_logdir),
            call(nagios_exportdir),
        ])
        os_.mkdir.assert_called_with(nagios_logdir)
        os_.chown.assert_called_with(nagios_logdir, nagios_uid, nagios_gid)
        self.assertEqual(os_.remove.mock_calls, [
            call(os.path.join(nagios_exportdir, 'foo-check_vhost.cfg')),
            call(os.path.join(nagios_exportdir, 'bar-check_vhost.cfg')),
        ])
        nrpe_service_config.write.assert_called_with(
            self.read_fixture('nrpe_service_config'))
        nrpe_check_config.write.assert_called_with(
            self.read_fixture('nrpe_check_config'))
        os_.path.isfile.assert_called_with('/etc/init.d/nagios-nrpe-server')
        call_.assert_called_with(['service', 'nagios-nrpe-server', 'reload'])

    @patch('hooks.config_get')
    @patch('hooks.log')
    def test_doesnt_update_nrpe_if_http_params_not_defined(self, log,
                                                           config_get):
        config_data = {}

        config_get.return_value = config_data

        hooks.update_nrpe_checks()

        log.assert_called_with("No vhost check data, exiting")

    @patch('hooks.config_get')
    @patch('pwd.getpwnam')
    @patch('grp.getgrnam')
    @patch('hooks.log')
    def test_doesnt_update_nrpe_if_no_uid(self, log, getpwnam, getgrnam,
                                          config_get):
        config_data = {
            'nagios_check_http_params': ['foo', 'bar'],
            'nagios_context': 'somecontext',
        }
        nagios_gid = 2345

        getpwnam.side_effect = Exception()
        getgrnam('nagios').gr_gid = nagios_gid

        config_get.return_value = config_data

        hooks.update_nrpe_checks()

        log.assert_called_with("Nagios user not setup, exiting")

    @patch('hooks.config_get')
    @patch('pwd.getpwnam')
    @patch('grp.getgrnam')
    @patch('hooks.log')
    def test_doesnt_update_nrpe_if_no_gid(self, log, getpwnam, getgrnam,
                                          config_get):
        config_data = {
            'nagios_check_http_params': ['foo', 'bar'],
            'nagios_context': 'somecontext',
        }
        nagios_uid = 1234

        getpwnam('nagios').pw_uid = nagios_uid
        getgrnam.side_effect = Exception()

        config_get.return_value = config_data

        hooks.update_nrpe_checks()

        log.assert_called_with("Nagios user not setup, exiting")

    @patch('hooks.config_get')
    @patch('pwd.getpwnam')
    @patch('grp.getgrnam')
    @patch('hooks.os')  # patch the reference that "hooks" has to "os"
    @patch('hooks.log')
    @patch('subprocess.call')
    def test_doesnt_update_nrpe_if_no_export_dir(self, call_, log, os_,
                                                 getgrnam, getpwnam,
                                                 config_get):
        config_data = {
            'nagios_check_http_params': ['foo', 'bar'],
            'nagios_context': 'somecontext',
        }
        os_.environ = {
            'JUJU_UNIT_NAME': 'some/unit',
        }
        nagios_logdir_exists = False
        nagios_exportdir_exists = False
        nagios_uid = 1234
        nagios_gid = 2345
        nagios_logdir = '/var/log/nagios'
        nagios_exportdir = '/var/lib/nagios/export'

        config_get.return_value = config_data
        getpwnam('nagios').pw_uid = nagios_uid
        getgrnam('nagios').gr_gid = nagios_gid
        os_.path.exists.side_effect = [
            nagios_logdir_exists,
            nagios_exportdir_exists,
        ]
        os_.path.join = os.path.join

        hooks.update_nrpe_checks()

        getpwnam.assert_called_with('nagios')
        getgrnam.assert_called_with('nagios')
        self.assertEqual(os_.path.exists.mock_calls, [
            call(nagios_logdir),
            call(nagios_exportdir),
        ])
        os_.mkdir.assert_called_with(nagios_logdir)
        os_.chown.assert_called_with(nagios_logdir, nagios_uid, nagios_gid)
        log.assert_called_with('Exiting as %s is not accessible' % (
            nagios_exportdir))
        self.assertFalse(call_.called)


class HooksTest(TestCase):
    def setUp(self):
        super(HooksTest, self).setUp()
        self.executable_file = '/some/executable'
        self.unexecutable_file = '/some/unexecutable'
        self.is_a_file = True
        self.not_a_file = False
        self.is_executable = True
        self.not_executable = False
        self.dir_exists = True
        self.not_a_dir = False
        self.log_prefix = 'baz'
        self.log_path = '/tmp/pprint-%s.log' % self.log_prefix

    def tearDown(self):
        super(HooksTest, self).tearDown()
        if os.path.exists(self.log_path):
            os.remove(self.log_path)

    @patch('subprocess.check_call')
    @patch('glob.glob')
    @patch('os.path.isfile')
    @patch('os.access')
    @patch('os.path.exists')
    @patch('os.mkdir')
    @patch('hooks.apt_get_install')
    def test_installs_hook(self, apt_get_install, mkdir, exists, access,
                           isfile, glob, check_call):
        glob.return_value = [self.executable_file, self.unexecutable_file]
        isfile.side_effect = [self.is_a_file, self.is_a_file]
        access.side_effect = [self.is_executable, self.not_executable]
        exists.return_value = self.not_a_dir
        apt_get_install.return_value = 'some result'

        result = hooks.install_hook()

        self.assertEqual(result, 'some result')
        glob.assert_called_with('exec.d/*/charm-pre-install')
        isfile.assert_has_calls([
            call(self.executable_file),
            call(self.unexecutable_file),
        ])
        access.assert_has_calls([
            call(self.executable_file, os.X_OK),
            call(self.unexecutable_file, os.X_OK),
        ])
        check_call.assert_has_calls([
            call(['sh', '-c', self.executable_file])
        ])
        exists.assert_called_with(hooks.default_apache2_service_config_dir)
        mkdir.assert_called_with(hooks.default_apache2_service_config_dir,
                                 0600)
        apt_get_install.assert_has_calls([
            call('python-jinja2'),
            call('apache2'),
        ])

    @patch('subprocess.check_call')
    @patch('glob.glob')
    @patch('os.path.isfile')
    @patch('os.access')
    @patch('os.path.exists')
    @patch('os.mkdir')
    @patch('hooks.apt_get_install')
    def test_doesnt_create_dir_to_install_hooks_if_not_needed(
            self, apt_get_install, mkdir, exists, access, isfile, glob,
            check_call):
        glob.return_value = [self.executable_file, self.unexecutable_file]
        isfile.side_effect = [self.is_a_file, self.is_a_file]
        access.side_effect = [self.is_executable, self.not_executable]
        exists.return_value = self.dir_exists
        apt_get_install.return_value = 'some result'

        result = hooks.install_hook()

        self.assertEqual(result, 'some result')
        glob.assert_called_with('exec.d/*/charm-pre-install')
        isfile.assert_has_calls([
            call(self.executable_file),
            call(self.unexecutable_file),
        ])
        access.assert_has_calls([
            call(self.executable_file, os.X_OK),
            call(self.unexecutable_file, os.X_OK),
        ])
        check_call.assert_has_calls([
            call(['sh', '-c', self.executable_file])
        ])
        exists.assert_called_with(hooks.default_apache2_service_config_dir)
        self.assertFalse(mkdir.called)
        apt_get_install.assert_has_calls([
            call('python-jinja2'),
            call('apache2'),
        ])

    def test_dumps_data_into_file(self):
        data = {'foo': 'bar'}

        hooks.dump_data(data, self.log_prefix)

        with open(self.log_path) as f:
            self.assertEqual(f.read().strip(), pformat(data).strip())

    def test_dumps_nothing_if_data_not_provided(self):
        data = None

        hooks.dump_data(data, self.log_prefix)

        self.assertFalse(os.path.exists(self.log_path))

    @patch('hooks.get_relation_data')
    @patch('yaml.safe_load')
    @patch('hooks.log')
    def test_gets_reverseproxy_data(self, log, load, get_relation_data):
        relation_data = {
            'foo-unit-1': {
                'port': 1234,
                'private-address': '10.11.12.13',
                'all_services': '/some/yaml/file',
            },
            'foo-unit-2': {
                'port': 1234,
                'private-address': '10.11.12.14',
                'all_services': '/some/yaml/file2',
            },
            'bar-unit-1': {
                'port': 3214,
                'private-address': '10.11.12.13',
                'all_services': '/some/yaml/file3',
            },
        }
        yaml_content = [
            {
                'service_name': 'some-service',
                'service_port': '2345',
            },
        ]

        get_relation_data.return_value = relation_data
        load.return_value = yaml_content

        result = hooks.get_reverseproxy_data(relation='baz-proxy')

        self.assertEqual(result, {
            'bar-unit-1': {'all_services': '/some/yaml/file3',
                           'port': 3214,
                           'private-address': '10.11.12.13'},
            'barunit': '10.11.12.13:3214',
            'barunit_all_services': '/some/yaml/file3',
            'barunit_port': 3214,
            'barunit_private-address': '10.11.12.13',
            'barunit_some-service': '10.11.12.13:2345',
            'foo-unit-1': {'all_services': '/some/yaml/file',
                           'port': 1234,
                           'private-address': '10.11.12.13'},
            'foo-unit-2': {'all_services': '/some/yaml/file2',
                           'port': 1234,
                           'private-address': '10.11.12.14'},
            'foounit': '10.11.12.13:1234',
            'foounit_all_services': '/some/yaml/file',
            'foounit_port': 1234,
            'foounit_private-address': '10.11.12.13',
            'foounit_some-service': '10.11.12.13:2345',
        })
        get_relation_data.assert_called_with(relation_name='baz-proxy')
        self.assertEqual(sorted(load.mock_calls), sorted([
            call('/some/yaml/file'),
            call('/some/yaml/file2'),
            call('/some/yaml/file3'),
        ]))
        self.assertEqual(sorted(log.mock_calls), sorted([
            call('unit_type: barunit'),
            call('unit_type: foounit'),
            call('unit_type: foounit'),
        ]))

    @patch('hooks.get_relation_data')
    @patch('yaml.safe_load')
    @patch('hooks.log')
    def test_truncates_reverseproxy_data_if_unit_has_no_port(
            self, log, load, get_relation_data):
        relation_data = {
            'foo-unit-1': {
                'port': 1234,
                'private-address': '10.11.12.13',
                'all_services': '/some/yaml/file',
            },
            'bar-unit-1': {
                'private-address': '10.11.12.13',
                'all_services': '/some/yaml/file3',
            },
        }

        get_relation_data.return_value = relation_data

        result = hooks.get_reverseproxy_data(relation='baz-proxy')

        self.assertEqual(result, {})
        get_relation_data.assert_called_with(relation_name='baz-proxy')
        self.assertFalse(load.called)

    @patch('hooks.get_relation_data')
    @patch('yaml.safe_load')
    @patch('hooks.log')
    def test_truncates_reverseproxy_data_if_no_data_available(
            self, log, load, get_relation_data):
        relation_data = {}

        get_relation_data.return_value = relation_data

        result = hooks.get_reverseproxy_data(relation='baz-proxy')

        self.assertEqual(result, {})
        get_relation_data.assert_called_with(relation_name='baz-proxy')
        self.assertFalse(load.called)


class RelationsTest(TestCase):
    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        call_args = ['relation-get', '--format=json']
        check_output.return_value = json_string

        result = hooks.relation_get()

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(call_args)
        log.assert_called_with('Calling: %s' % call_args)

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_with_scope(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        call_args = ['relation-get', '--format=json', 'baz-scope']
        check_output.return_value = json_string

        result = hooks.relation_get(scope='baz-scope')

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(call_args)
        log.assert_called_with('Calling: %s' % call_args)

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_with_unit_name(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        call_args = ['relation-get', '--format=json', 'baz-scope', '-',
                     'baz-unit']
        check_output.return_value = json_string

        result = hooks.relation_get(scope='baz-scope', unit_name='baz-unit')

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(call_args)
        log.assert_called_with('Calling: %s' % call_args)

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_with_relation_id(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        call_args = ['relation-get', '--format=json', '-r', 123, 'baz-scope',
                     '-', 'baz-unit']
        check_output.return_value = json_string

        result = hooks.relation_get(scope='baz-scope', unit_name='baz-unit',
                                    relation_id=123)

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(call_args)
        log.assert_called_with('Calling: %s' % call_args)

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_logs_and_returns_none_relation_get_fails(self, log,
                                                      check_output):
        check_output.side_effect = RuntimeError('some error')

        result = hooks.relation_get()

        log.assert_called_with('Calling: %s' % ['relation-get',
                                                '--format=json'])
        self.assertIsNone(result)

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_ids(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        check_output.return_value = json_string

        result = hooks.get_relation_ids()

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(['relation-ids', '--format=json'])
        log.assert_called_with('Calling: %s' % ['relation-ids',
                                                '--format=json'])

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_ids_by_name(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        check_output.return_value = json_string

        result = hooks.get_relation_ids(relation_name='baz')

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(['relation-ids', '--format=json',
                                         'baz'])
        log.assert_called_with('Calling: %s' % ['relation-ids',
                                                '--format=json', 'baz'])

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_returns_none_when_get_relation_ids_fails(self, log,
                                                      check_output):
        check_output.side_effect = RuntimeError('some error')

        result = hooks.get_relation_ids()

        log.assert_called_with('Calling: %s' % ['relation-ids',
                                                '--format=json'])
        self.assertIsNone(result)

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_list(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        check_output.return_value = json_string

        result = hooks.get_relation_list()

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(['relation-list', '--format=json'])
        log.assert_called_with('Calling: %s' % ['relation-list',
                                                '--format=json'])

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_gets_relation_list_by_id(self, log, check_output):
        json_string = '{"foo": "BAR"}'
        check_output.return_value = json_string

        result = hooks.get_relation_list(relation_id=123)

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(['relation-list', '--format=json',
                                         '-r', 123])
        log.assert_called_with('Calling: %s' % ['relation-list',
                                                '--format=json', '-r', 123])

    @patch('subprocess.check_output')
    @patch('hooks.log')
    def test_returns_none_when_get_relation_list_fails(self, log,
                                                       check_output):
        check_output.side_effect = RuntimeError('some error')

        result = hooks.get_relation_list()

        log.assert_called_with('Calling: %s' % ['relation-list',
                                                '--format=json'])
        self.assertIsNone(result)

    @patch('hooks.get_relation_ids')
    @patch('hooks.get_relation_list')
    @patch('hooks.relation_get')
    def test_gets_relation_data_by_name(self, relation_get, get_relation_list,
                                        get_relation_ids):
        get_relation_ids.return_value = [1, 2]
        get_relation_list.side_effect = [
            ['foo/1', 'bar/1'],
            ['foo/2', 'bar/2'],
        ]
        relation_get.side_effect = [
            'FOO 1',
            'BAR 1',
            'FOO 2',
            'BAR 2',
        ]

        result = hooks.get_relation_data(relation_name='baz')
        expected_data = {
            'foo-1': 'FOO 1',
            'bar-1': 'BAR 1',
            'foo-2': 'FOO 2',
            'bar-2': 'BAR 2',
        }

        self.assertEqual(result, expected_data)
        get_relation_ids.assert_called_with('baz')
        self.assertEqual(get_relation_list.mock_calls, [
            call(relation_id=1),
            call(relation_id=2),
        ])
        self.assertEqual(relation_get.mock_calls, [
            call(relation_id=1, unit_name='foo/1'),
            call(relation_id=1, unit_name='bar/1'),
            call(relation_id=2, unit_name='foo/2'),
            call(relation_id=2, unit_name='bar/2'),
        ])

    @patch('hooks.get_relation_ids')
    def test_gets_data_as_none_if_no_relation_ids_exist(self,
                                                        get_relation_ids):
        get_relation_ids.return_value = None

        result = hooks.get_relation_data(relation_name='baz')

        self.assertEqual(result, ())
        get_relation_ids.assert_called_with('baz')

    @patch('hooks.get_relation_ids')
    @patch('hooks.get_relation_list')
    def test_returns_none_if_get_data_fails(self, get_relation_list,
                                            get_relation_ids):
        get_relation_ids.return_value = [1, 2]
        get_relation_list.side_effect = RuntimeError('some error')

        result = hooks.get_relation_data(relation_name='baz')

        self.assertIsNone(result)
